@@ -7,17 +7,19 @@ import os
 import sys
 import itertools
 import traceback
+import textwrap
+import subprocess
 
 import numpy as np
 from numpy import array, single, double, csingle, cdouble, dot, identity
 from numpy import multiply, atleast_2d, inf, asarray, matrix
 from numpy import linalg
-from numpy.linalg import matrix_power, norm, matrix_rank, multi_dot
+from numpy.linalg import matrix_power, norm, matrix_rank, multi_dot, LinAlgError
 from numpy.linalg.linalg import _multi_dot_matrix_chain_order
 from numpy.testing import (
     assert_, assert_equal, assert_raises, assert_array_equal,
     assert_almost_equal, assert_allclose, run_module_suite,
-    dec
+    dec, SkipTest, suppress_warnings
 )
 
 
@@ -32,11 +34,11 @@ def imply(a, b):
 old_assert_almost_equal = assert_almost_equal
 
 
-def assert_almost_equal(a, b, **kw):
+def assert_almost_equal(a, b, single_decimal=6, double_decimal=12, **kw):
     if asarray(a).dtype.type in (single, csingle):
-        decimal = 6
+        decimal = single_decimal
     else:
-        decimal = 12
+        decimal = double_decimal
     old_assert_almost_equal(a, b, decimal=decimal, **kw)
 
 
@@ -58,24 +60,43 @@ def get_rtol(dtype):
         return 1e-11
 
 
-class LinalgCase(object):
+# used to categorize tests
+all_tags = {
+  'square', 'nonsquare', 'hermitian',  # mutually exclusive
+  'generalized', 'size-0', 'strided' # optional additions
+}
 
-    def __init__(self, name, a, b, exception_cls=None):
-        assert isinstance(name, str)
+class LinalgCase(object):
+    def __init__(self, name, a, b, tags=set()):
+        """
+        A bundle of arguments to be passed to a test case, with an identifying
+        name, the operands a and b, and a set of tags to filter the tests
+        """
+        assert_(isinstance(name, str))
         self.name = name
         self.a = a
         self.b = b
-        self.exception_cls = exception_cls
+        self.tags = frozenset(tags)  # prevent shared tags
 
     def check(self, do):
-        if self.exception_cls is None:
-            do(self.a, self.b)
-        else:
-            assert_raises(self.exception_cls, do, self.a, self.b)
+        """
+        Run the function `do` on this test case, expanding arguments
+        """
+        do(self.a, self.b, tags=self.tags)
 
     def __repr__(self):
         return "<LinalgCase: %s>" % (self.name,)
 
+def apply_tag(tag, cases):
+    """
+    Add the given tag (a string) to each of the cases (a list of LinalgCase
+    objects)
+    """
+    assert tag in all_tags, "Invalid tag"
+    for case in cases:
+        case.tags = case.tags | {tag}
+    return cases
+
 
 #
 # Base test cases
@@ -83,7 +104,10 @@ class LinalgCase(object):
 
 np.random.seed(1234)
 
-SQUARE_CASES = [
+CASES = []
+
+# square test cases
+CASES += apply_tag('square', [
     LinalgCase("single",
                array([[1., 2.], [3., 4.]], dtype=single),
                array([2., 1.], dtype=single)),
@@ -102,10 +126,14 @@ SQUARE_CASES = [
     LinalgCase("cdouble_2",
                array([[1. + 2j, 2 + 3j], [3 + 4j, 4 + 5j]], dtype=cdouble),
                array([[2. + 1j, 1. + 2j, 1 + 3j], [1 - 2j, 1 - 3j, 1 - 6j]], dtype=cdouble)),
-    LinalgCase("empty",
-               atleast_2d(array([], dtype=double)),
-               atleast_2d(array([], dtype=double)),
-               linalg.LinAlgError),
+    LinalgCase("0x0",
+               np.empty((0, 0), dtype=double),
+               np.empty((0,), dtype=double),
+               tags={'size-0'}),
+    LinalgCase("0x0_matrix",
+               np.empty((0, 0), dtype=double).view(np.matrix),
+               np.empty((0, 1), dtype=double).view(np.matrix),
+               tags={'size-0'}),
     LinalgCase("8x8",
                np.random.rand(8, 8),
                np.random.rand(8)),
@@ -121,9 +149,10 @@ SQUARE_CASES = [
     LinalgCase("matrix_a_and_b",
                matrix([[1., 2.], [3., 4.]]),
                matrix([2., 1.]).T),
-]
+])
 
-NONSQUARE_CASES = [
+# non-square test-cases
+CASES += apply_tag('nonsquare', [
     LinalgCase("single_nsq_1",
                array([[1., 2., 3.], [3., 4., 6.]], dtype=single),
                array([2., 1.], dtype=single)),
@@ -162,16 +191,25 @@ NONSQUARE_CASES = [
                array([[2. + 1j, 1. + 2j], [1 - 1j, 2 - 2j], [1 - 1j, 2 - 2j]], dtype=cdouble)),
     LinalgCase("8x11",
                np.random.rand(8, 11),
-               np.random.rand(11)),
+               np.random.rand(8)),
     LinalgCase("1x5",
                np.random.rand(1, 5),
-               np.random.rand(5)),
+               np.random.rand(1)),
     LinalgCase("5x1",
                np.random.rand(5, 1),
-               np.random.rand(1)),
-]
+               np.random.rand(5)),
+    LinalgCase("0x4",
+               np.random.rand(0, 4),
+               np.random.rand(0),
+               tags={'size-0'}),
+    LinalgCase("4x0",
+               np.random.rand(4, 0),
+               np.random.rand(4),
+               tags={'size-0'}),
+])
 
-HERMITIAN_CASES = [
+# hermitian test-cases
+CASES += apply_tag('hermitian', [
     LinalgCase("hsingle",
                array([[1., 2.], [2., 1.]], dtype=single),
                None),
@@ -185,9 +223,9 @@ HERMITIAN_CASES = [
                array([[1., 2 + 3j], [2 - 3j, 1]], dtype=cdouble),
                None),
     LinalgCase("hempty",
-               atleast_2d(array([], dtype=double)),
+               np.empty((0, 0), dtype=double),
                None,
-               linalg.LinAlgError),
+               tags={'size-0'}),
     LinalgCase("hnonarray",
                [[1, 2], [2, 1]],
                None),
@@ -200,21 +238,16 @@ HERMITIAN_CASES = [
     LinalgCase("hmatrix_1x1",
                np.random.rand(1, 1),
                None),
-]
+])
 
 
 #
 # Gufunc test cases
 #
+def _make_generalized_cases():
+    new_cases = []
 
-GENERALIZED_SQUARE_CASES = []
-GENERALIZED_NONSQUARE_CASES = []
-GENERALIZED_HERMITIAN_CASES = []
-
-for tgt, src in ((GENERALIZED_SQUARE_CASES, SQUARE_CASES),
-                 (GENERALIZED_NONSQUARE_CASES, NONSQUARE_CASES),
-                 (GENERALIZED_HERMITIAN_CASES, HERMITIAN_CASES)):
-    for case in src:
+    for case in CASES:
         if not isinstance(case.a, np.ndarray):
             continue
 
@@ -224,8 +257,8 @@ for tgt, src in ((GENERALIZED_SQUARE_CASES, SQUARE_CASES),
         else:
             b = np.array([case.b, 7 * case.b, 6 * case.b])
         new_case = LinalgCase(case.name + "_tile3", a, b,
-                              case.exception_cls)
-        tgt.append(new_case)
+                              tags=case.tags | {'generalized'})
+        new_cases.append(new_case)
 
         a = np.array([case.a] * 2 * 3).reshape((3, 2) + case.a.shape)
         if case.b is None:
@@ -233,14 +266,17 @@ for tgt, src in ((GENERALIZED_SQUARE_CASES, SQUARE_CASES),
         else:
             b = np.array([case.b] * 2 * 3).reshape((3, 2) + case.b.shape)
         new_case = LinalgCase(case.name + "_tile213", a, b,
-                              case.exception_cls)
-        tgt.append(new_case)
+                              tags=case.tags | {'generalized'})
+        new_cases.append(new_case)
+
+    return new_cases
+
+CASES += _make_generalized_cases()
 
 #
 # Generate stride combination variations of the above
 #
 
-
 def _stride_comb_iter(x):
     """
     Generate cartesian product of strides for all axes
@@ -267,7 +303,7 @@ def _stride_comb_iter(x):
         xi = xi[slices]
         xi[...] = x
         xi = xi.view(x.__class__)
-        assert np.all(xi == x)
+        assert_(np.all(xi == x))
         yield xi, "stride_" + "_".join(["%+d" % j for j in repeats])
 
         # generate also zero strides if possible
@@ -288,29 +324,35 @@ def _stride_comb_iter(x):
             xi = np.lib.stride_tricks.as_strided(x, strides=s)
             yield xi, "stride_xxx_0_0"
 
-for src in (SQUARE_CASES,
-            NONSQUARE_CASES,
-            HERMITIAN_CASES,
-            GENERALIZED_SQUARE_CASES,
-            GENERALIZED_NONSQUARE_CASES,
-            GENERALIZED_HERMITIAN_CASES):
-
+def _make_strided_cases():
     new_cases = []
-    for case in src:
-        for a, a_tag in _stride_comb_iter(case.a):
-            for b, b_tag in _stride_comb_iter(case.b):
-                new_case = LinalgCase(case.name + "_" + a_tag + "_" + b_tag, a, b,
-                                      exception_cls=case.exception_cls)
+    for case in CASES:
+        for a, a_label in _stride_comb_iter(case.a):
+            for b, b_label in _stride_comb_iter(case.b):
+                new_case = LinalgCase(case.name + "_" + a_label + "_" + b_label, a, b,
+                                      tags=case.tags | {'strided'})
                 new_cases.append(new_case)
-    src.extend(new_cases)
+    return new_cases
+
+CASES += _make_strided_cases()
 
 
 #
 # Test different routines against the above cases
 #
 
-def _check_cases(func, cases):
-    for case in cases:
+def _check_cases(func, require=set(), exclude=set()):
+    """
+    Run func on each of the cases with all of the tags in require, and none
+    of the tags in exclude
+    """
+    for case in CASES:
+        # filter by require and exclude
+        if case.tags & require != require:
+            continue
+        if case.tags & exclude:
+            continue
+
         try:
             case.check(func)
         except Exception:
@@ -319,43 +361,67 @@ def _check_cases(func, cases):
             raise AssertionError(msg)
 
 
-class LinalgTestCase(object):
+class LinalgSquareTestCase(object):
 
     def test_sq_cases(self):
-        _check_cases(self.do, SQUARE_CASES)
+        _check_cases(self.do, require={'square'}, exclude={'generalized', 'size-0'})
+
+    def test_empty_sq_cases(self):
+        _check_cases(self.do, require={'square', 'size-0'}, exclude={'generalized'})
 
 
 class LinalgNonsquareTestCase(object):
 
-    def test_sq_cases(self):
-        _check_cases(self.do, NONSQUARE_CASES)
+    def test_nonsq_cases(self):
+        _check_cases(self.do, require={'nonsquare'}, exclude={'generalized', 'size-0'})
+
+    def test_empty_nonsq_cases(self):
+        _check_cases(self.do, require={'nonsquare', 'size-0'}, exclude={'generalized'})
+
+class HermitianTestCase(object):
+
+    def test_herm_cases(self):
+        _check_cases(self.do, require={'hermitian'}, exclude={'generalized', 'size-0'})
+
+    def test_empty_herm_cases(self):
+        _check_cases(self.do, require={'hermitian', 'size-0'}, exclude={'generalized'})
 
 
-class LinalgGeneralizedTestCase(object):
+class LinalgGeneralizedSquareTestCase(object):
 
     @dec.slow
     def test_generalized_sq_cases(self):
-        _check_cases(self.do, GENERALIZED_SQUARE_CASES)
+        _check_cases(self.do, require={'generalized', 'square'}, exclude={'size-0'})
+
+    @dec.slow
+    def test_generalized_empty_sq_cases(self):
+        _check_cases(self.do, require={'generalized', 'square', 'size-0'})
 
 
 class LinalgGeneralizedNonsquareTestCase(object):
 
     @dec.slow
     def test_generalized_nonsq_cases(self):
-        _check_cases(self.do, GENERALIZED_NONSQUARE_CASES)
+        _check_cases(self.do, require={'generalized', 'nonsquare'}, exclude={'size-0'})
 
-
-class HermitianTestCase(object):
-
-    def test_herm_cases(self):
-        _check_cases(self.do, HERMITIAN_CASES)
+    @dec.slow
+    def test_generalized_empty_nonsq_cases(self):
+        _check_cases(self.do, require={'generalized', 'nonsquare', 'size-0'})
 
 
 class HermitianGeneralizedTestCase(object):
 
     @dec.slow
     def test_generalized_herm_cases(self):
-        _check_cases(self.do, GENERALIZED_HERMITIAN_CASES)
+        _check_cases(self.do,
+            require={'generalized', 'hermitian'},
+            exclude={'size-0'})
+
+    @dec.slow
+    def test_generalized_empty_herm_cases(self):
+        _check_cases(self.do,
+            require={'generalized', 'hermitian', 'size-0'},
+            exclude={'none'})
 
 
 def dot_generalized(a, b):
@@ -388,9 +454,9 @@ def identity_like_generalized(a):
         return identity(a.shape[0])
 
 
-class TestSolve(LinalgTestCase, LinalgGeneralizedTestCase):
+class TestSolve(LinalgSquareTestCase, LinalgGeneralizedSquareTestCase):
 
-    def do(self, a, b):
+    def do(self, a, b, tags):
         x = linalg.solve(a, b)
         assert_almost_equal(b, dot_generalized(a, x))
         assert_(imply(isinstance(b, matrix), isinstance(x, matrix)))
@@ -454,9 +520,9 @@ class TestSolve(LinalgTestCase, LinalgGeneralizedTestCase):
         assert_(isinstance(result, ArraySubclass))
 
 
-class TestInv(LinalgTestCase, LinalgGeneralizedTestCase):
+class TestInv(LinalgSquareTestCase, LinalgGeneralizedSquareTestCase):
 
-    def do(self, a, b):
+    def do(self, a, b, tags):
         a_inv = linalg.inv(a)
         assert_almost_equal(dot_generalized(a, a_inv),
                             identity_like_generalized(a))
@@ -477,17 +543,18 @@ class TestInv(LinalgTestCase, LinalgGeneralizedTestCase):
         res = linalg.inv(a)
         assert_(res.dtype.type is np.float64)
         assert_equal(a.shape, res.shape)
-        assert_(isinstance(a, ArraySubclass))
+        assert_(isinstance(res, ArraySubclass))
 
         a = np.zeros((0, 0), dtype=np.complex64).view(ArraySubclass)
         res = linalg.inv(a)
         assert_(res.dtype.type is np.complex64)
         assert_equal(a.shape, res.shape)
+        assert_(isinstance(res, ArraySubclass))
 
 
-class TestEigvals(LinalgTestCase, LinalgGeneralizedTestCase):
+class TestEigvals(LinalgSquareTestCase, LinalgGeneralizedSquareTestCase):
 
-    def do(self, a, b):
+    def do(self, a, b, tags):
         ev = linalg.eigvals(a)
         evalues, evectors = linalg.eig(a)
         assert_almost_equal(ev, evalues)
@@ -501,10 +568,28 @@ class TestEigvals(LinalgTestCase, LinalgGeneralizedTestCase):
         for dtype in [single, double, csingle, cdouble]:
             yield check, dtype
 
+    def test_0_size(self):
+        # Check that all kinds of 0-sized arrays work
+        class ArraySubclass(np.ndarray):
+            pass
+        a = np.zeros((0, 1, 1), dtype=np.int_).view(ArraySubclass)
+        res = linalg.eigvals(a)
+        assert_(res.dtype.type is np.float64)
+        assert_equal((0, 1), res.shape)
+        # This is just for documentation, it might make sense to change:
+        assert_(isinstance(res, np.ndarray))
 
-class TestEig(LinalgTestCase, LinalgGeneralizedTestCase):
+        a = np.zeros((0, 0), dtype=np.complex64).view(ArraySubclass)
+        res = linalg.eigvals(a)
+        assert_(res.dtype.type is np.complex64)
+        assert_equal((0,), res.shape)
+        # This is just for documentation, it might make sense to change:
+        assert_(isinstance(res, np.ndarray))
 
-    def do(self, a, b):
+
+class TestEig(LinalgSquareTestCase, LinalgGeneralizedSquareTestCase):
+
+    def do(self, a, b, tags):
         evalues, evectors = linalg.eig(a)
         assert_allclose(dot_generalized(a, evectors),
                         np.asarray(evectors) * np.asarray(evalues)[..., None, :],
@@ -526,10 +611,36 @@ class TestEig(LinalgTestCase, LinalgGeneralizedTestCase):
         for dtype in [single, double, csingle, cdouble]:
             yield check, dtype
 
+    def test_0_size(self):
+        # Check that all kinds of 0-sized arrays work
+        class ArraySubclass(np.ndarray):
+            pass
+        a = np.zeros((0, 1, 1), dtype=np.int_).view(ArraySubclass)
+        res, res_v = linalg.eig(a)
+        assert_(res_v.dtype.type is np.float64)
+        assert_(res.dtype.type is np.float64)
+        assert_equal(a.shape, res_v.shape)
+        assert_equal((0, 1), res.shape)
+        # This is just for documentation, it might make sense to change:
+        assert_(isinstance(a, np.ndarray))
 
-class TestSVD(LinalgTestCase, LinalgGeneralizedTestCase):
+        a = np.zeros((0, 0), dtype=np.complex64).view(ArraySubclass)
+        res, res_v = linalg.eig(a)
+        assert_(res_v.dtype.type is np.complex64)
+        assert_(res.dtype.type is np.complex64)
+        assert_equal(a.shape, res_v.shape)
+        assert_equal((0,), res.shape)
+        # This is just for documentation, it might make sense to change:
+        assert_(isinstance(a, np.ndarray))
+
+
+class TestSVD(LinalgSquareTestCase, LinalgGeneralizedSquareTestCase):
+
+    def do(self, a, b, tags):
+        if 'size-0' in tags:
+            assert_raises(LinAlgError, linalg.svd, a, 0)
+            return
 
-    def do(self, a, b):
         u, s, vt = linalg.svd(a, 0)
         assert_allclose(a, dot_generalized(np.asarray(u) * np.asarray(s)[..., None, :],
                                            np.asarray(vt)),
@@ -550,27 +661,45 @@ class TestSVD(LinalgTestCase, LinalgGeneralizedTestCase):
         for dtype in [single, double, csingle, cdouble]:
             yield check, dtype
 
+    def test_0_size(self):
+        # These raise errors currently
+        # (which does not mean that it may not make sense)
+        a = np.zeros((0, 0), dtype=np.complex64)
+        assert_raises(linalg.LinAlgError, linalg.svd, a)
+        a = np.zeros((0, 1), dtype=np.complex64)
+        assert_raises(linalg.LinAlgError, linalg.svd, a)
+        a = np.zeros((1, 0), dtype=np.complex64)
+        assert_raises(linalg.LinAlgError, linalg.svd, a)
 
-class TestCondSVD(LinalgTestCase, LinalgGeneralizedTestCase):
 
-    def do(self, a, b):
+class TestCondSVD(LinalgSquareTestCase, LinalgGeneralizedSquareTestCase):
+
+    def do(self, a, b, tags):
         c = asarray(a)  # a might be a matrix
+        if 'size-0' in tags:
+            assert_raises(LinAlgError, linalg.svd, c, compute_uv=False)
+            return
         s = linalg.svd(c, compute_uv=False)
-        old_assert_almost_equal(
-            s[..., 0] / s[..., -1], linalg.cond(a), decimal=5)
+        assert_almost_equal(
+            s[..., 0] / s[..., -1], linalg.cond(a),
+            single_decimal=5, double_decimal=11)
 
     def test_stacked_arrays_explicitly(self):
         A = np.array([[1., 2., 1.], [0, -2., 0], [6., 2., 3.]])
         assert_equal(linalg.cond(A), linalg.cond(A[None, ...])[0])
 
 
-class TestCond2(LinalgTestCase):
+class TestCond2(LinalgSquareTestCase):
 
-    def do(self, a, b):
+    def do(self, a, b, tags):
         c = asarray(a)  # a might be a matrix
+        if 'size-0' in tags:
+            assert_raises(LinAlgError, linalg.svd, c, compute_uv=False)
+            return
         s = linalg.svd(c, compute_uv=False)
-        old_assert_almost_equal(
-            s[..., 0] / s[..., -1], linalg.cond(a, 2), decimal=5)
+        assert_almost_equal(
+            s[..., 0] / s[..., -1], linalg.cond(a, 2),
+            single_decimal=5, double_decimal=11)
 
     def test_stacked_arrays_explicitly(self):
         A = np.array([[1., 2., 1.], [0, -2., 0], [6., 2., 3.]])
@@ -584,17 +713,22 @@ class TestCondInf(object):
         assert_almost_equal(linalg.cond(A, inf), 3.)
 
 
-class TestPinv(LinalgTestCase):
+class TestPinv(LinalgSquareTestCase,
+               LinalgNonsquareTestCase,
+               LinalgGeneralizedSquareTestCase,
+               LinalgGeneralizedNonsquareTestCase):
 
-    def do(self, a, b):
+    def do(self, a, b, tags):
         a_ginv = linalg.pinv(a)
-        assert_almost_equal(dot(a, a_ginv), identity(asarray(a).shape[0]))
+        # `a @ a_ginv == I` does not hold if a is singular
+        dot = dot_generalized
+        assert_almost_equal(dot(dot(a, a_ginv), a), a, single_decimal=5, double_decimal=11)
         assert_(imply(isinstance(a, matrix), isinstance(a_ginv, matrix)))
 
 
-class TestDet(LinalgTestCase, LinalgGeneralizedTestCase):
+class TestDet(LinalgSquareTestCase, LinalgGeneralizedSquareTestCase):
 
-    def do(self, a, b):
+    def do(self, a, b, tags):
         d = linalg.det(a)
         (s, ld) = linalg.slogdet(a)
         if asarray(a).dtype.type in (single, double):
@@ -634,14 +768,37 @@ class TestDet(LinalgTestCase, LinalgGeneralizedTestCase):
         for dtype in [single, double, csingle, cdouble]:
             yield check, dtype
 
+    def test_0_size(self):
+        a = np.zeros((0, 0), dtype=np.complex64)
+        res = linalg.det(a)
+        assert_equal(res, 1.)
+        assert_(res.dtype.type is np.complex64)
+        res = linalg.slogdet(a)
+        assert_equal(res, (1, 0))
+        assert_(res[0].dtype.type is np.complex64)
+        assert_(res[1].dtype.type is np.float32)
 
-class TestLstsq(LinalgTestCase, LinalgNonsquareTestCase):
+        a = np.zeros((0, 0), dtype=np.float64)
+        res = linalg.det(a)
+        assert_equal(res, 1.)
+        assert_(res.dtype.type is np.float64)
+        res = linalg.slogdet(a)
+        assert_equal(res, (1, 0))
+        assert_(res[0].dtype.type is np.float64)
+        assert_(res[1].dtype.type is np.float64)
+
+
+class TestLstsq(LinalgSquareTestCase, LinalgNonsquareTestCase):
+
+    def do(self, a, b, tags):
+        if 'size-0' in tags:
+            assert_raises(LinAlgError, linalg.lstsq, a, b)
+            return
 
-    def do(self, a, b):
         arr = np.asarray(a)
         m, n = arr.shape
         u, s, vt = linalg.svd(a, 0)
-        x, residuals, rank, sv = linalg.lstsq(a, b)
+        x, residuals, rank, sv = linalg.lstsq(a, b, rcond=-1)
         if m <= n:
             assert_almost_equal(b, dot(a, x))
             assert_equal(rank, m)
@@ -652,7 +809,7 @@ class TestLstsq(LinalgTestCase, LinalgNonsquareTestCase):
             expect_resids = (
                 np.asarray(abs(np.dot(a, x) - b)) ** 2).sum(axis=0)
             expect_resids = np.asarray(expect_resids)
-            if len(np.asarray(b).shape) == 1:
+            if np.asarray(b).ndim == 1:
                 expect_resids.shape = (1,)
                 assert_equal(residuals.shape, expect_resids.shape)
         else:
@@ -662,6 +819,23 @@ class TestLstsq(LinalgTestCase, LinalgNonsquareTestCase):
         assert_(imply(isinstance(b, matrix), isinstance(x, matrix)))
         assert_(imply(isinstance(b, matrix), isinstance(residuals, matrix)))
 
+    def test_future_rcond(self):
+        a = np.array([[0., 1.,  0.,  1.,  2.,  0.],
+                      [0., 2.,  0.,  0.,  1.,  0.],
+                      [1., 0.,  1.,  0.,  0.,  4.],
+                      [0., 0.,  0.,  2.,  3.,  0.]]).T
+
+        b = np.array([1, 0, 0, 0, 0, 0])
+        with suppress_warnings() as sup:
+            w = sup.record(FutureWarning, "`rcond` parameter will change")
+            x, residuals, rank, s = linalg.lstsq(a, b)
+            assert_(rank == 4)
+            x, residuals, rank, s = linalg.lstsq(a, b, rcond=-1)
+            assert_(rank == 4)
+            x, residuals, rank, s = linalg.lstsq(a, b, rcond=None)
+            assert_(rank == 3)
+            # Warning should be raised exactly once (first command)
+            assert_(len(w) == 1)
 
 class TestMatrixPower(object):
     R90 = array([[0, 1], [-1, 0]])
@@ -728,7 +902,7 @@ class TestBoolPower(object):
 
 class TestEigvalsh(HermitianTestCase, HermitianGeneralizedTestCase):
 
-    def do(self, a, b):
+    def do(self, a, b, tags):
         # note that eigenvalue arrays returned by eig must be sorted since
         # their order isn't guaranteed.
         ev = linalg.eigvalsh(a, 'L')
@@ -775,10 +949,28 @@ class TestEigvalsh(HermitianTestCase, HermitianGeneralizedTestCase):
         w = np.linalg.eigvalsh(Kup, UPLO='u')
         assert_allclose(w, tgt, rtol=rtol)
 
+    def test_0_size(self):
+        # Check that all kinds of 0-sized arrays work
+        class ArraySubclass(np.ndarray):
+            pass
+        a = np.zeros((0, 1, 1), dtype=np.int_).view(ArraySubclass)
+        res = linalg.eigvalsh(a)
+        assert_(res.dtype.type is np.float64)
+        assert_equal((0, 1), res.shape)
+        # This is just for documentation, it might make sense to change:
+        assert_(isinstance(res, np.ndarray))
+
+        a = np.zeros((0, 0), dtype=np.complex64).view(ArraySubclass)
+        res = linalg.eigvalsh(a)
+        assert_(res.dtype.type is np.float32)
+        assert_equal((0,), res.shape)
+        # This is just for documentation, it might make sense to change:
+        assert_(isinstance(res, np.ndarray))
+
 
 class TestEigh(HermitianTestCase, HermitianGeneralizedTestCase):
 
-    def do(self, a, b):
+    def do(self, a, b, tags):
         # note that eigenvalue arrays returned by eig must be sorted since
         # their order isn't guaranteed.
         ev, evc = linalg.eigh(a)
@@ -834,6 +1026,28 @@ class TestEigh(HermitianTestCase, HermitianGeneralizedTestCase):
         w, v = np.linalg.eigh(Kup, UPLO='u')
         assert_allclose(w, tgt, rtol=rtol)
 
+    def test_0_size(self):
+        # Check that all kinds of 0-sized arrays work
+        class ArraySubclass(np.ndarray):
+            pass
+        a = np.zeros((0, 1, 1), dtype=np.int_).view(ArraySubclass)
+        res, res_v = linalg.eigh(a)
+        assert_(res_v.dtype.type is np.float64)
+        assert_(res.dtype.type is np.float64)
+        assert_equal(a.shape, res_v.shape)
+        assert_equal((0, 1), res.shape)
+        # This is just for documentation, it might make sense to change:
+        assert_(isinstance(a, np.ndarray))
+
+        a = np.zeros((0, 0), dtype=np.complex64).view(ArraySubclass)
+        res, res_v = linalg.eigh(a)
+        assert_(res_v.dtype.type is np.complex64)
+        assert_(res.dtype.type is np.float32)
+        assert_equal(a.shape, res_v.shape)
+        assert_equal((0,), res.shape)
+        # This is just for documentation, it might make sense to change:
+        assert_(isinstance(a, np.ndarray))
+
 
 class _TestNorm(object):
 
@@ -845,6 +1059,98 @@ class _TestNorm(object):
         assert_equal(norm(array([], dtype=self.dt)), 0.0)
         assert_equal(norm(atleast_2d(array([], dtype=self.dt))), 0.0)
 
+    def test_vector_return_type(self):
+        a = np.array([1, 0, 1])
+
+        exact_types = np.typecodes['AllInteger']
+        inexact_types = np.typecodes['AllFloat']
+
+        all_types = exact_types + inexact_types
+
+        for each_inexact_types in all_types:
+            at = a.astype(each_inexact_types)
+
+            an = norm(at, -np.inf)
+            assert_(issubclass(an.dtype.type, np.floating))
+            assert_almost_equal(an, 0.0)
+
+            with suppress_warnings() as sup:
+                sup.filter(RuntimeWarning, "divide by zero encountered")
+                an = norm(at, -1)
+                assert_(issubclass(an.dtype.type, np.floating))
+                assert_almost_equal(an, 0.0)
+
+            an = norm(at, 0)
+            assert_(issubclass(an.dtype.type, np.floating))
+            assert_almost_equal(an, 2)
+
+            an = norm(at, 1)
+            assert_(issubclass(an.dtype.type, np.floating))
+            assert_almost_equal(an, 2.0)
+
+            an = norm(at, 2)
+            assert_(issubclass(an.dtype.type, np.floating))
+            assert_almost_equal(an, an.dtype.type(2.0)**an.dtype.type(1.0/2.0))
+
+            an = norm(at, 4)
+            assert_(issubclass(an.dtype.type, np.floating))
+            assert_almost_equal(an, an.dtype.type(2.0)**an.dtype.type(1.0/4.0))
+
+            an = norm(at, np.inf)
+            assert_(issubclass(an.dtype.type, np.floating))
+            assert_almost_equal(an, 1.0)
+
+    def test_matrix_return_type(self):
+        a = np.array([[1, 0, 1], [0, 1, 1]])
+
+        exact_types = np.typecodes['AllInteger']
+
+        # float32, complex64, float64, complex128 types are the only types
+        # allowed by `linalg`, which performs the matrix operations used
+        # within `norm`.
+        inexact_types = 'fdFD'
+
+        all_types = exact_types + inexact_types
+
+        for each_inexact_types in all_types:
+            at = a.astype(each_inexact_types)
+
+            an = norm(at, -np.inf)
+            assert_(issubclass(an.dtype.type, np.floating))
+            assert_almost_equal(an, 2.0)
+
+            with suppress_warnings() as sup:
+                sup.filter(RuntimeWarning, "divide by zero encountered")
+                an = norm(at, -1)
+                assert_(issubclass(an.dtype.type, np.floating))
+                assert_almost_equal(an, 1.0)
+
+            an = norm(at, 1)
+            assert_(issubclass(an.dtype.type, np.floating))
+            assert_almost_equal(an, 2.0)
+
+            an = norm(at, 2)
+            assert_(issubclass(an.dtype.type, np.floating))
+            assert_almost_equal(an, 3.0**(1.0/2.0))
+
+            an = norm(at, -2)
+            assert_(issubclass(an.dtype.type, np.floating))
+            assert_almost_equal(an, 1.0)
+
+            an = norm(at, np.inf)
+            assert_(issubclass(an.dtype.type, np.floating))
+            assert_almost_equal(an, 2.0)
+
+            an = norm(at, 'fro')
+            assert_(issubclass(an.dtype.type, np.floating))
+            assert_almost_equal(an, 2.0)
+
+            an = norm(at, 'nuc')
+            assert_(issubclass(an.dtype.type, np.floating))
+            # Lower bar needed to support low precision floats.
+            # They end up being off by 1 in the 7th place.
+            old_assert_almost_equal(an, 2.7320508075688772, decimal=6)
+
     def test_vector(self):
         a = [1, 2, 3, 4]
         b = [-1, -2, -3, -4]
@@ -1009,8 +1315,8 @@ class _TestNorm(object):
             assert_raises(ValueError, norm, B, order, (1, 2))
 
         # Invalid axis
-        assert_raises(ValueError, norm, B, None, 3)
-        assert_raises(ValueError, norm, B, None, (2, 3))
+        assert_raises(np.AxisError, norm, B, None, 3)
+        assert_raises(np.AxisError, norm, B, None, (2, 3))
         assert_raises(ValueError, norm, B, None, (0, 1, 2))
 
 
@@ -1072,11 +1378,25 @@ class TestMatrixRank(object):
         yield assert_equal, matrix_rank(np.zeros((4,))), 0
         # accepts array-like
         yield assert_equal, matrix_rank([1]), 1
-        # greater than 2 dimensions raises error
-        yield assert_raises, TypeError, matrix_rank, np.zeros((2, 2, 2))
+        # greater than 2 dimensions treated as stacked matrices
+        ms = np.array([I, np.eye(4), np.zeros((4,4))])
+        yield assert_equal, matrix_rank(ms), np.array([3, 4, 0])
         # works on scalar
         yield assert_equal, matrix_rank(1), 1
 
+    def test_symmetric_rank(self):
+        yield assert_equal, 4, matrix_rank(np.eye(4), hermitian=True)
+        yield assert_equal, 1, matrix_rank(np.ones((4, 4)), hermitian=True)
+        yield assert_equal, 0, matrix_rank(np.zeros((4, 4)), hermitian=True)
+        # rank deficient matrix
+        I = np.eye(4)
+        I[-1, -1] = 0.
+        yield assert_equal, 3, matrix_rank(I, hermitian=True)
+        # manually supplied tolerance
+        I[-1, -1] = 1e-8
+        yield assert_equal, 4, matrix_rank(I, hermitian=True, tol=0.99e-8)
+        yield assert_equal, 3, matrix_rank(I, hermitian=True, tol=1.01e-8)
+
 
 def test_reduced_rank():
     # Test matrices with reduced rank
@@ -1175,6 +1495,59 @@ class TestQR(object):
             self.check_qr(m2.T)
             self.check_qr(matrix(m1))
 
+    def test_0_size(self):
+        # There may be good ways to do (some of this) reasonably:
+        a = np.zeros((0, 0))
+        assert_raises(linalg.LinAlgError, linalg.qr, a)
+        a = np.zeros((0, 1))
+        assert_raises(linalg.LinAlgError, linalg.qr, a)
+        a = np.zeros((1, 0))
+        assert_raises(linalg.LinAlgError, linalg.qr, a)
+
+
+class TestCholesky(object):
+    # TODO: are there no other tests for cholesky?
+
+    def test_basic_property(self):
+        # Check A = L L^H
+        shapes = [(1, 1), (2, 2), (3, 3), (50, 50), (3, 10, 10)]
+        dtypes = (np.float32, np.float64, np.complex64, np.complex128)
+
+        for shape, dtype in itertools.product(shapes, dtypes):
+            np.random.seed(1)
+            a = np.random.randn(*shape)
+            if np.issubdtype(dtype, np.complexfloating):
+                a = a + 1j*np.random.randn(*shape)
+
+            t = list(range(len(shape)))
+            t[-2:] = -1, -2
+
+            a = np.matmul(a.transpose(t).conj(), a)
+            a = np.asarray(a, dtype=dtype)
+
+            c = np.linalg.cholesky(a)
+
+            b = np.matmul(c, c.transpose(t).conj())
+            assert_allclose(b, a,
+                            err_msg="{} {}\n{}\n{}".format(shape, dtype, a, c),
+                            atol=500 * a.shape[0] * np.finfo(dtype).eps)
+
+    def test_0_size(self):
+        class ArraySubclass(np.ndarray):
+            pass
+        a = np.zeros((0, 1, 1), dtype=np.int_).view(ArraySubclass)
+        res = linalg.cholesky(a)
+        assert_equal(a.shape, res.shape)
+        assert_(res.dtype.type is np.float64)
+        # for documentation purpose:
+        assert_(isinstance(res, np.ndarray))
+
+        a = np.zeros((1, 0, 0), dtype=np.complex64).view(ArraySubclass)
+        res = linalg.cholesky(a)
+        assert_equal(a.shape, res.shape)
+        assert_(res.dtype.type is np.complex64)
+        assert_(isinstance(res, np.ndarray))
+
 
 def test_byteorder_check():
     # Byte order check should pass for native order
@@ -1215,7 +1588,6 @@ def test_xerbla_override():
     # Check that our xerbla has been successfully linked in. If it is not,
     # the default xerbla routine is called, which prints a message to stdout
     # and may, or may not, abort the process depending on the LAPACK package.
-    from nose import SkipTest
 
     XERBLA_OK = 255
 
@@ -1237,7 +1609,7 @@ def test_xerbla_override():
             np.linalg.lapack_lite.xerbla()
         except ValueError:
             pass
-        except:
+        except Exception:
             os._exit(os.EX_CONFIG)
 
         try:
@@ -1261,6 +1633,40 @@ def test_xerbla_override():
             raise SkipTest('Numpy xerbla not linked in.')
 
 
+def test_sdot_bug_8577():
+    # Regression test that loading certain other libraries does not
+    # result to wrong results in float32 linear algebra.
+    #
+    # There's a bug gh-8577 on OSX that can trigger this, and perhaps
+    # there are also other situations in which it occurs.
+    #
+    # Do the check in a separate process.
+
+    bad_libs = ['PyQt5.QtWidgets', 'IPython']
+
+    template = textwrap.dedent("""
+    import sys
+    {before}
+    try:
+        import {bad_lib}
+    except ImportError:
+        sys.exit(0)
+    {after}
+    x = np.ones(2, dtype=np.float32)
+    sys.exit(0 if np.allclose(x.dot(x), 2.0) else 1)
+    """)
+
+    for bad_lib in bad_libs:
+        code = template.format(before="import numpy as np", after="",
+                               bad_lib=bad_lib)
+        subprocess.check_call([sys.executable, "-c", code])
+
+        # Swapped import order
+        code = template.format(after="import numpy as np", before="",
+                               bad_lib=bad_lib)
+        subprocess.check_call([sys.executable, "-c", code])
+
+
 class TestMultiDot(object):
 
     def test_basic_function_with_three_arguments(self):
@@ -1332,7 +1738,7 @@ class TestMultiDot(object):
                                [0,  0,  0,  3,  3,  3],
                                [0,  0,  0,  0,  4,  5],
                                [0,  0,  0,  0,  0,  5],
-                               [0,  0,  0,  0,  0,  0]], dtype=np.int)
+                               [0,  0,  0,  0,  0,  0]], dtype=int)
         s_expected -= 1  # Cormen uses 1-based index, python does not.
 
         s, m = _multi_dot_matrix_chain_order(arrays, return_costs=True)
